@@ -44,7 +44,8 @@ enum _GtkHTMLEditPropertyType {
 
 GtkHTMLEditPropertiesDialog * gtk_html_edit_properties_dialog_new          (GtkHTMLControlData *cd,
 									    gboolean insert,
-									    gchar *title);
+									    gchar *title,
+									    gchar *icon_path);
 void                          gtk_html_edit_properties_dialog_destroy      (GtkHTMLEditPropertiesDialog *d);
 
 void                          gtk_html_edit_properties_dialog_add_entry    (GtkHTMLEditPropertiesDialog *d,
