@@ -70,21 +70,28 @@ normalize (HTMLObject **object,
 }
 
 
+
+inline void
+html_cursor_init (HTMLCursor *cursor, HTMLObject *o, guint offset)
+{
+	cursor->object = o;
+	cursor->offset = offset;
+
+	cursor->target_x = 0;
+	cursor->have_target_x = FALSE;
+
+	cursor->position = 0;
+}
+
 HTMLCursor *
 html_cursor_new (void)
 {
-	HTMLCursor *new;
+	HTMLCursor *new_cursor;
 
-	new = g_new (HTMLCursor, 1);
-	new->object = NULL;
-	new->offset = 0;
+	new_cursor = g_new (HTMLCursor, 1);
+	html_cursor_init (new_cursor, NULL, 0);
 
-	new->target_x = 0;
-	new->have_target_x = FALSE;
-
-	new->position = 0;
-
-	return new;
+	return new_cursor;
 }
 
 void
@@ -137,22 +144,6 @@ html_cursor_normalize (HTMLCursor *cursor)
 	normalize (&cursor->object, &cursor->offset);
 }
 
-
-/* This is a gross hack as we don't have a `is_a()' function in the object
-   system.  */
-
-static gboolean
-is_clue (HTMLObject *object)
-{
-	HTMLType type;
-
-	type = HTML_OBJECT_TYPE (object);
-
-	return (type == HTML_TYPE_CLUE || type == HTML_TYPE_CLUEV
-		|| type == HTML_TYPE_CLUEH || type == HTML_TYPE_CLUEFLOW);
-}
-
-
 void
 html_cursor_home (HTMLCursor *cursor,
 		  HTMLEngine *engine)
@@ -172,11 +163,20 @@ html_cursor_home (HTMLCursor *cursor,
 		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
 
 	obj = engine->clue;
-	while (is_clue (obj))
-		obj = HTML_CLUE (obj)->head;
+	while (!html_object_accepts_cursor (obj)) {
+		HTMLObject *head = html_object_head (obj);
+		if (obj)
+			obj = head;
+		else
+			break;
+	}
 
 	cursor->object = obj;
 	cursor->offset = 0;
+
+	if (!html_object_accepts_cursor (obj))
+		html_cursor_forward (cursor, engine);
+
 	cursor->position = 0;
 
 	debug_location (cursor);
