@@ -15,8 +15,7 @@
 
     You should have received a copy of the GNU Library General Public License
     along with this library; see the file COPYING.LIB.  If not, write to
-    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-    Boston, MA 02111-1307, USA.
+    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,    Boston, MA 02111-1307, USA.
 
 */
 
@@ -37,7 +36,7 @@ html_form_new (HTMLEngine *engine, gchar *_action, gchar *_method)
 
 	new->elements = NULL;
 	new->hidden = NULL;
-	new->engine = engine;
+	html_form_set_engine (new, engine);
 
 	new->radio_group = g_hash_table_new (g_str_hash, g_str_equal);
 
@@ -147,6 +146,13 @@ html_form_submit (HTMLForm *form)
 	g_string_free (encoding, TRUE);
 }
 
+void
+html_form_set_engine (HTMLForm *form, HTMLEngine *engine)
+{
+	g_return_if_fail (HTML_IS_ENGINE (engine));
+	form->engine = engine;
+}
+
 void
 html_form_reset (HTMLForm *form)
 {
